import numpy as np
from scipy import signal as s
import pandas as pd
from scipy.optimize import leastsq

class SignalAnalysis:

    @staticmethod
    def findRowsWithExactPatternInSeries(series, pattern):
        seriesWithLastIndexOfPattern = series.rolling(len(pattern)).apply(lambda x: all(np.equal(x, pattern)),
                                                                          raw=False)
        dfWithPattern = seriesWithLastIndexOfPattern.fillna(0)
        lastIndicesOfPattern = np.where(dfWithPattern)[0]  # indices with the last found element
        allIndicesOfPattern = [range(match - len(pattern) + 1, match + 1) for match in
                               lastIndicesOfPattern]  # all indices corresponding to the pattern

        data = []
        for subs in allIndicesOfPattern:
            data.append(series.iloc[subs].index)

        return pd.DataFrame(data, columns=pattern)

    @staticmethod
    def findRowsWithExactPatternInSeries(series, pattern):
        seriesWithLastIndexOfPattern = series.rolling(len(pattern)).apply(lambda x: all(np.equal(x, pattern)),
                                                                          raw=False)
        dfWithPattern = seriesWithLastIndexOfPattern.fillna(0)
        lastIndicesOfPattern = np.where(dfWithPattern)[0]  # indices with the last found element
        allIndicesOfPattern = [range(match - len(pattern) + 1, match + 1) for match in
                               lastIndicesOfPattern]  # all indices corresponding to the pattern

        data = []
        for subs in allIndicesOfPattern:
            data.append(series.iloc[subs].index)

        return pd.DataFrame(data, columns=pattern)
    
    @staticmethod
    def resampleSeries(originSeries, ts=0):
        """Function resamples series by interpolating in order to get continuous sample frequency"""
        # takes sample time of first sample if not given
        if ts == 0:
            ts = (originSeries.index[1] - originSeries.index[0])  # Sample distance
        T = (originSeries.index[-1] - originSeries.index[0])  # Timeperiod
        N = int(T / ts)  # Number of samplepoints
        indexList = np.linspace(0, T, N)

        # interpolate data in order to have the same sample time everywhere
        interpolatedSeries = originSeries.reindex(originSeries.index.union(indexList)).interpolate('index', limit_direction='forward')  # akima

        # delete those points who are not in in the continous intervall
        delIndex = [x for x in originSeries.index if x not in indexList]
        interpolatedSeries = interpolatedSeries.drop(delIndex, axis=0)

        return interpolatedSeries
    
    @staticmethod
    def getWindowWithGivenMeanAndStd(originSeries, indexIncrement=20, mean=50, std=0.1):
        """Method returns a series within a given mean and standart deviation"""
        for j in range(len(originSeries) - indexIncrement):
            window = originSeries.iloc[j:j + indexIncrement]
            windowMean = window.mean()
            windowStd = window.std()
            isWindowFound = windowMean >= mean and windowStd <= std

            if isWindowFound:
                minIndex = originSeries.iloc[j:].idxmin() # find the time index of minimum value
                minIndex = originSeries.index.get_loc(minIndex) # find the iteration of this index
                return originSeries.iloc[j:minIndex].copy()

        # throw exception if not found
        raise Error("No window found")
        
    @staticmethod
    def lowpassFilter(originSeries, fs=0, fc=0):
        """Method applies low pass filter to series"""
        #Set cuttoff frequency if not given
        if fs == 0:
            fs = 1/(originSeries.index[1]-originSeries.index[0]) #Sample distance
        if fc == 0:
            fc = fs/100  # Cut-off frequency of the filter
        w = fc / (fs / 2) # Normalize the frequency

        b, a = s.butter(5, w, 'low')
        filteredSeries = pd.DataFrame(data = s.filtfilt(b, a, originSeries.values), index = [originSeries.index], columns = ['filtered'])
        return filteredSeries['filtered']
    
    @staticmethod
    def getWindowWithDecayOnly(originSeries, initialValue, deviationAllowance=0.996):
        """Method applies low pass filter to series"""   
        # use filter to find start of decay
        filteredSeries = SignalAnalysis.lowpassFilter(originSeries)
        # get Df with initial values
        initialSeries = filteredSeries[filteredSeries.values > initialValue*deviationAllowance]
        # Only take values after the initial phase 
        decaySeries = originSeries[originSeries.index > initialSeries.index[-1]]
        return decaySeries

    def calculateTimeConstantOfExponentialDecayInSecondsWithIntegral(exponentialDecayInSeconds, exponent):
        return exponent * np.trapz(exponentialDecayInSeconds, x=exponentialDecayInSeconds.index, axis=0) /(exponentialDecayInSeconds.iloc[0] - exponentialDecayInSeconds.iloc[-1])

    def chargeApproach(sliceDF):
        exponent=1
        sliceSquaredDF = sliceDF.pow(exponent)
        return SignalAnalysis.calculateTimeConstantOfExponentialDecayInSecondsWithIntegral(sliceSquaredDF, exponent)

    def energyApproach(sliceDF):
        exponent=2
        sliceSquaredDF = sliceDF.pow(exponent)
        return SignalAnalysis.calculateTimeConstantOfExponentialDecayInSecondsWithIntegral(sliceSquaredDF, exponent)

    def linReg(myDF):
        right_exp = lambda p, x: -p[1]*x+p[0]
        err = lambda p, x, y:(right_exp(p, x) -y)

        time = myDF.index
        rate = np.log(myDF.values)    
        v0= [0., time.min() ]
        out = leastsq(err, v0, args = (time, rate))
        v = out[0] #fit parameters out []
        xxx = np.arange(min(time), max(time), time[1] - time[0])
        ccc = right_exp(v, xxx)
        timeConstant = 1/v[1]
        return timeConstant

    def eFit(myDF):
        right_exp = lambda p, x: p[0]*np.exp(-p[2]*(x-p[1]))
        err = lambda p, x, y:(right_exp(p, x) -y)
        
        time = myDF.index
        rate = myDF.values
        v0= [0., 0., time.min() ]
        out = leastsq(err, v0, args = (time, rate))
        v = out[0] #fit parameters out []
        xxx = np.arange(min(time), max(time), time[1] - time[0])
        ccc = right_exp(v, xxx)
        timeConstant = 1/v[2]
        return timeConstant

    __nameToFunction = {"chargeApproach": chargeApproach, "energyApproach": energyApproach, "linReg": linReg, "eFit": eFit}

    @staticmethod
    def getDecayTimeConstant(df, functionNamesToApply):
        """
               
            **returns dataframe with the time constants, calculated with the chosen methods**

            Function returns a dataframe with time constant according to the input method chosen for the passed DataFrame object
            Note that passed DataFrame object should just contain the decay, the functions getWindowWithDecayOnly or
            getWindowWithGivenMeanAndStd can be used for filtering

            :type df: DF 
            :type functionNamesToApply: list
            :return: DF

            - Example :

            from source.signals.SignalAnalysis import SignalAnalysis
            import pandas as pd
            import numpy as np

            x = np.arange(100)*0.001
            timeConstant = 0.8
            df = pd.DataFrame(data = np.exp(-x/timeConstant), index = x, columns = ["Voltage_1"])
            
            functionNamesToApply = ["chargeApproach", "energyApproach", "linReg", "eFit"]
            SignalAnalysis.getDecayTimeConstant(df, functionNamesToApply)
            
            >>> 					Voltage_1
            >>>chargeApproach		0.8
            >>>energyApproach		0.8
            >>>linReg				0.8
            >>>eFit					0.8
            
        """

        functionsToApply = [SignalAnalysis.__nameToFunction[x] for x in functionNamesToApply]
        if isinstance(df, pd.DataFrame) and not df.empty:
            return df.aggregate(functionsToApply)
        else:
            return df
    
    
    @staticmethod
    def characteristicOfExponentialDecay(originSeries,window=100):
        """Method returns the change in the characteristic of the exponential decay over time as a series"""
        samplingtime= (originSeries.index[1]-originSeries.index[0])
        difSeries = originSeries.diff(periods=window)/samplingtime/window
        tauSeries = -originSeries.div(difSeries)
        return tauSeries
    
    @staticmethod
    def spectrogramFromSeries(originSeries):
        """Function calculates spectrogram from DF"""

        originSeries = originSeries.to_frame(name=None).set_index((originSeries.index-originSeries.index[0]))[originSeries.name]  #*10**-9
        ts = (originSeries.index[1]-originSeries.index[0]) #Sample distance
        T = (originSeries.index[-1]-originSeries.index[0]) #Timeperiod
        N = int(T/ts) # Number of samplepoints
        if len(originSeries) == N: #are the samples continuous
            newSeries = originSeries
        else:
            newSeries = SignalAnalysis.resampleSeries(originSeries)
        nperseg = N // 128
        f, t, Sxx = s.spectrogram(newSeries.values, 1/ts ,window=('hann'), nperseg=nperseg, noverlap = nperseg // 2)

        return f, t, Sxx
    
    @staticmethod
    def filterBankFromSpectrogram(Sxx):
        """Function filterbank feature from sprectrogram"""
        filterBankMaxValues = np.zeros(len(Sxx[:,1])//8)
        for i in range(len(filterBankMaxValues)):
            k = i*8           
            filterBankMaxValues[i] = np.amax(Sxx[k:k+8,:])

        return(filterBankMaxValues)
        
    @staticmethod     
    def crossCorrelation(refSeries, measSeries):
        """Function compares to signals and returns the correlation coefficient, the length of the signals can vary"""
        shift = len(measSeries.index) - len(refSeries.index)
        if shift != 0:
            R_val = np.zeros(shift);
            i_val = np.zeros(shift);
            for i in range(0 ,abs(shift)):
                i_val[i] += i;
                R = np.corrcoef(measSeries.values[i:i+len(refSeries.values)], refSeries.values);
                R_val[i] = R[0][1];

            if shift >0:   
                maxCorrelationTimestamp = measSeries.index[np.argmax(R_val)]
            else:
                maxCorrelationTimestamp = refSeries.index[np.argmax(R_val)]
            maxCorrelation = np.max(R_val)
            print("Maximum correlation is: " + str(maxCorrelation) + " at: " + str(maxCorrelationTimestamp))
        else:
            maxCorrelation = np.corrcoef(measSeries.values, refSeries.values);
            print("Maximum correlation is: " + str(maxCorrelation))

        return maxCorrelation    


